// -*-Java-*-
//
//  This file is part of the Siena Simulator, a simple discrete-event
//  simulator used with the Siena project.  See
//  http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//          Matthew J. Rutherford <rutherfo@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2003 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: ProcessWithPId.java,v 1.3 2003/04/22 21:26:44 mjr Exp $
//
package ssim;

/**  utility Process class providing a utility interface with the
 *  simulator.
 *
 *  This is a sligtly more advanced Process class that provides 
 *  automatic management of the process-id.
 **/
public abstract class ProcessWithPId implements Process
{
    private long m_pid;

    public ProcessWithPId()
    {
	m_pid = NULL_PROCESSID;
    }

    /**  activates this process within the simulator.
     *
     *  Creates a simulator process with this process object.  The
     *  same ProcessWithPId can be activated only once.
     *
     *  @return the ProcessId of the created simulator process. Or
     *  NULL_PROCESSID if this process object is already associated
     *  with a simulation process.
     *
     *  @see Sim#create_process(Process,short) Sim.create_process(Process,short)
     **/
    public long activate()
    {
	return activate( (short)0 );
    }

    public long activate( short mode )
    {
	if( m_pid == NULL_PROCESSID )
	    {
		return m_pid = Sim.create_process( this, mode );
	    }
	else
	    {
		return NULL_PROCESSID;
	    }
    }
    
    /**  process id of this process.
     *
     *  @return the id of the simulation process with this objectk,
     *  or NULL_PROCESSID if no process is associated with this
     *  object.
     **/
    public long pid()
    {
	return m_pid;
    }

    public void process_timeout()
    {
    }

    public void stop()
    {
    }
}
