package tests;

import java.util.*;
import ssim.*;

public class bs extends junit.framework.TestCase
{
    public final static int ORDER_EVENT_TYPE = 1;
    public final static int DELIVER_EVENT_TYPE = 2;
    public final static Random RAND = new Random();

    static class BSEvent extends Event
    {
	public BSEvent( int type )
	{
	    super( type );
	}

	public String toString()
	{
	    return "";
	}
    }

    static class Order extends BSEvent
    {
	String m_item;
	int m_orderId;
	int m_quantity;
	long m_deliverToId;

	public Order( String item, int quantity, long deliverToId )
	{
	    super( ORDER_EVENT_TYPE );
	    m_item = item;
	    m_orderId = RAND.nextInt();
	    m_quantity = quantity;
	    m_deliverToId = deliverToId;
	}

	public Order( Order ord )
	{
	    super( ORDER_EVENT_TYPE );
	    m_item = ord.m_item;
	    m_orderId = ord.m_orderId;
	    m_quantity = ord.m_quantity;
	    m_deliverToId = ord.m_deliverToId;
	}

	public String getItem()
	{
	    return m_item;
	}

	public int getOrderId()
	{
	    return m_orderId;
	}

	public int getQuantity()
	{
	    return m_quantity;
	}

	public long getDeliverToId()
	{
	    return m_deliverToId;
	}

	public String toString()
	{
	    return "[Order: type=" + type +
			",item=" + m_item +
			",orderId=" + m_orderId +
			",quantity=" + m_quantity +
			",deliverToId=" + m_deliverToId +
			"]";
	}
    }

    static class Delivery extends BSEvent
    {
	String m_item;
	int m_orderId;
	int m_quantity;

	public Delivery( String item, int orderId, int quantity )
	{
	    super( DELIVER_EVENT_TYPE );
	    m_item = item;
	    m_orderId = orderId;
	    m_quantity = quantity;
	}

	public String getItem()
	{
	    return m_item;
	}

	public int getOrderId()
	{
	    return m_orderId;
	}

	public int getQuantity()
	{
	    return m_quantity;
	}

	public String toString()
	{
	    return "[Delivery: type=" + type +
			",item=" + m_item +
			",orderId=" + m_orderId +
			",quantity=" + m_quantity +
			"]";
	}
    }

    static class Manufacturer extends ProcessWithPId
    {
	HashMap m_delayMap = new HashMap();

	public Manufacturer()
	{
	}

	public void addItem( String item, int delay )
	{
	    System.out.println( "Manufacturer(" + pid() + ") -> " + item + " takes " + delay + " days to make." );
	    m_delayMap.put( item, new Integer( delay ) );
	}

	public void process_event( Event msg )
	{
	    if( msg.type == ORDER_EVENT_TYPE ) {
		handleOrder( msg );
	    }
	    else {
		System.err.println( "Manufacturer(" + pid() + ") received unknown event: " + msg );
	    }
	}

	private void handleOrder( Event msg )
	{
	    Order ord = (Order)msg;
	    System.out.println( "Manufacturer(" + pid() + ") received: " + ord );
	    int delay = ((Integer)m_delayMap.get( ord.getItem() )).intValue();
	    Delivery del = new Delivery( ord.getItem(), ord.getOrderId(), ord.getQuantity() );
	    // add a random extra delay
	    Sim.signal_event( del, ord.getDeliverToId(), delay + RAND.nextInt( 10 ) );
	}

	public void init()
	{
	    System.out.println( "Manufacturer: init" );
	    Sim.set_timeout( 0 );
	}
    }

    static class item_count
    {
	int stock;
	int held;
	int ordered;
	int claimed;
    };

    class Retailer extends ProcessWithPId
    {
	long m_mid;
	HashMap m_stockMap = new HashMap();
	HashMap m_pendingOrders = new HashMap();

	public Retailer( long mid )
	{
	    m_mid = mid;
	}

	public void process_event( Event evt )
	{
	    if( evt.type == DELIVER_EVENT_TYPE ) {
		handleDelivery( evt );
	    }
	    else if( evt.type == ORDER_EVENT_TYPE ) {
		handleOrder( evt );
	    }
	    else {
		System.err.println( "ERROR: Retailer(" + pid() + ") received unknown event: " + evt );
	    }
	}

	public void addItem( String item, int quantity )
	{
	    System.out.println( "Retailer(" + pid() +  ") -> has " + quantity + " " + item + "(s)" );
	    item_count ic = new item_count();
	    ic.stock = quantity;
	    ic.held = 0;
	    ic.ordered = 0;
	    ic.claimed = 0;
	    m_stockMap.put( item, ic );
	}

	private void handleDelivery( Event msg )
	{
	    Delivery del = (Delivery)msg;
	    System.out.println( "Retailer(" + pid() +  ") received: " + del );
	    item_count ic = (item_count)m_stockMap.get( del.getItem() );
	    // a delivery of X widgets arrived, we need to combine with
	    // the widgets that are being held and start sending deliveries
	    // out to the buyer.
	    ic.ordered -= del.getQuantity();
	    int q_delivered = del.getQuantity();
	    
	    Vector oq = (Vector)m_pendingOrders.get( del.getItem() );
	    Vector newQ = new Vector();
	    while( !oq.isEmpty() ) {
		Order ord = (Order)oq.firstElement();
		oq.removeElementAt( 0 );

		int q_needed = ord.getQuantity();
		if( (ic.stock + ic.held + q_delivered) >= q_needed ) {
		    // we are going to fulfill this order one way or the other...
		    Delivery del2 = new Delivery( ord.getItem(),
						 ord.getOrderId(),
						 ord.getQuantity() );
		    Sim.signal_event( del2, ord.getDeliverToId(), 1 );
		    // just have to account for it, first from held
		    q_needed -= ic.held;
		    if( q_needed < 0 ) {
			ic.held = -1 * q_needed;
			q_needed = 0;
		    }
		    else {
			ic.held = 0;
		    }
		    if( q_needed == 0 ) {
			break;
		    }
		    // second from stock
		    q_needed -= ic.stock;
		    if( q_needed < 0 ) {
			ic.stock = -1 * q_needed;
			q_needed = 0;
		    }
		    else {
			ic.stock = 0;
		    }
		    if( q_needed == 0 ) {
			break;
		    }
		    // third from delivery
		    if( q_delivered < q_needed ) {
			System.err.println( "LOGIC ERROR: q_needed=" + q_needed );
			break;
		    }
		    q_delivered -= q_needed;
		    ic.claimed -= q_needed;
		}
		else {
		    // we can't fulfill the order yet
		    newQ.add( ord );
		}
	    }
	    m_pendingOrders.put( del.getItem(), newQ );

	    // if q_delivered is still positive, it goes into stock
	    ic.stock += q_delivered;
	}
	
	void handleOrder( Event msg )
	{
	    Order ord = (Order)msg;
	    System.out.println( "Retailer(" + pid() + ") received: " + ord );
	    /*
	     * an order was received by this retailer, it needs to 
	     * check if it can fulfill the order out of stock. If
	     * not, it needs to order the difference, and put
	     * the order in a holding area until it gets delivery.
	     */
	    int q_needed = ord.getQuantity();
	    item_count ic = (item_count)m_stockMap.get( ord.getItem() );
	    if( ic.stock >= q_needed ) {
		    Delivery del = new Delivery( ord.getItem(),
						 ord.getOrderId(),
						 q_needed );
		    Sim.signal_event( del, ord.getDeliverToId(), 1 );
		    ic.stock -=  q_needed;
	    }
	    else {
		// move the ones in stock to held
		q_needed -= ic.stock;
		ic.held += ic.stock;
		ic.stock = 0;

		// need to wait for a delivery, have to copy
		// order because scheduler may delete order after this 
		// method returns.
		Vector q = (Vector)m_pendingOrders.get( ord.getItem() );
		if( q == null ) {
		    q = new Vector();
		    m_pendingOrders.put( ord.getItem(), q );
		}
		q.add( new Order( ord ) );
		ic.claimed += q_needed;
		// if there aren't already enough on order, generate an order 
		if( (ic.ordered - ic.claimed) < 0 ) {
		    int q2 = ic.claimed - ic.ordered;
		    Order ord2 = new Order( ord.getItem(), q2, pid() );
		    Sim.signal_event( ord2, m_mid, 1 );
		    ic.ordered += q2;
		}
	    }
	}

	public void init()
	{
	    System.out.println( "Retailer: init" );
	    Sim.set_timeout( 0 );
	}
    }

    static class Buyer extends ProcessWithPId
    {
	long m_rid;
	HashMap m_shoppingMap = new HashMap();
	HashMap m_delayMap =  new HashMap();

	public Buyer()
	{
	    m_rid = ssim.Process.NULL_PROCESSID;
	}

	public void process_event( Event evt )
	{
	    if( evt.type == DELIVER_EVENT_TYPE ) {
		handleDelivery( evt );
	    }
	    else {
		System.err.println( "ERROR: Buyer(" +  pid() + ") received unknown event: " +  evt );
	    }
	}

	public void addItem( String item, int quantity )
	{
	    m_shoppingMap.put( item, new Integer( quantity ) );
	}

	private void handleDelivery( Event msg )
	{
	    Delivery del = (Delivery)msg;
	    int sendDelay = ((Integer)m_delayMap.get( del.getItem() )).intValue();
	    System.out.println( "Buyer(" + pid() +  ") order for " + del.getQuantity() + " " + del.getItem() +
				" took " + (Sim.clock()-sendDelay) + " days." );
	}

	public void init()
	{
	    System.out.println( "Buyer.init()" );
	    Iterator itr = m_shoppingMap.entrySet().iterator();
	    while( itr.hasNext() ) {
		Map.Entry i = (Map.Entry)itr.next();
		Order ord = new Order( (String)i.getKey(), ((Integer)i.getValue()).intValue(), pid() );
		int delay = RAND.nextInt( 20 );
		m_delayMap.put( i.getKey(), new Integer( delay ) );
		Sim.signal_event( ord, m_rid, delay );
	    }
	}

	public void setRetailerId( long id )
	{
	    m_rid = id;
	}
    }

    private final static int BUYER_COUNT = 10;

    public bs( String name )
    {
	super( name );
    }

    public void test()
    {
	Random rand = new Random( 1 );

	Set items = new HashSet();
	items.add( "HardDrive" );
	items.add( "Keyboard" );
	items.add( "Monitor" );
	items.add( "Mouse" );
	items.add( "Printer" );
	items.add( "Scanner" );

	Manufacturer m = new Manufacturer();
	m.activate();

	Retailer r = new Retailer( m.pid() );
	r.activate();

	Buyer b[]  = new Buyer [ BUYER_COUNT ];

	Iterator i = items.iterator();
	boolean begin = true;
	while( i.hasNext() ) {
	    String item = (String)i.next();
	    // name, production delay
	    m.addItem( item, (1 + rand.nextInt( 29 )) );
	    
	    // name, quantity in stock
	    r.addItem( item, (1 + rand.nextInt( 99 )) );

	    for( int j=0; j<BUYER_COUNT; j++ ) {
		if( begin ){
		    b[j] = new Buyer();
		    b[j].activate();
		    b[j].setRetailerId( r.pid() );
		}
		// name, number needed
		b[j].addItem( item, (1 + rand.nextInt( 49 )) );
	    }
	    begin = false;
	}

	Sim.run_simulation();
    }
}
